# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from types import SimpleNamespace
from typing import Any, Dict, Optional

from playwright._impl._api_structures import DebuggerLocation, DebuggerPausedDetails
from playwright._impl._connection import ChannelOwner


class Debugger(ChannelOwner):
    Events = SimpleNamespace(
        PausedStateChanged="pausedstatechanged",
    )

    def __init__(
        self, parent: ChannelOwner, type: str, guid: str, initializer: Dict
    ) -> None:
        super().__init__(parent, type, guid, initializer)
        self._paused_details: Optional[DebuggerPausedDetails] = None
        self._channel.on(
            "pausedStateChanged", lambda params: self._on_paused_state_changed(params)
        )

    def _on_paused_state_changed(self, params: Dict[str, Any]) -> None:
        self._paused_details = params.get("pausedDetails")
        self.emit(Debugger.Events.PausedStateChanged)

    async def request_pause(self) -> None:
        await self._channel.send("requestPause", None)

    async def resume(self) -> None:
        await self._channel.send("resume", None)

    async def next(self) -> None:
        await self._channel.send("next", None)

    async def run_to(self, location: DebuggerLocation) -> None:
        await self._channel.send("runTo", None, {"location": location})

    @property
    def paused_details(self) -> Optional[DebuggerPausedDetails]:
        return self._paused_details



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import inspect
import traceback
from typing import Awaitable, Callable, Dict

import greenlet

from playwright._impl._connection import ChannelOwner
from playwright._impl._errors import Error, is_target_closed_error


class Disposable(ChannelOwner):
    def __init__(
        self, parent: ChannelOwner, type: str, guid: str, initializer: Dict
    ) -> None:
        super().__init__(parent, type, guid, initializer)

    async def dispose(self) -> None:
        try:
            await self._channel.send(
                "dispose",
                None,
            )
        except Exception as e:
            if not is_target_closed_error(e):
                raise e

    async def close(self) -> None:
        await self.dispose()

    def __repr__(self) -> str:
        return "<Disposable>"


class DisposableStub:
    def __init__(
        self,
        dispose_fn: Callable[[], Awaitable[None]],
        parent: ChannelOwner,
    ) -> None:
        self._dispose_fn = dispose_fn
        self._loop = parent._loop
        self._dispatcher_fiber = parent._dispatcher_fiber

    async def dispose(self) -> None:
        await self._dispose_fn()

    async def __aenter__(self) -> "DisposableStub":
        return self

    async def __aexit__(self, *args: object) -> None:
        await self.dispose()

    def __enter__(self) -> "DisposableStub":
        return self

    def __exit__(self, *args: object) -> None:
        self._sync(self.dispose())

    def _sync(self, coro: object) -> object:
        __tracebackhide__ = True
        if self._loop.is_closed():
            coro.close()  # type: ignore
            raise Error("Event loop is closed! Is Playwright already stopped?")
        g_self = greenlet.getcurrent()
        task = self._loop.create_task(coro)  # type: ignore
        setattr(task, "__pw_stack__", inspect.stack(0))
        setattr(task, "__pw_stack_trace__", traceback.extract_stack(limit=10))
        task.add_done_callback(lambda _: g_self.switch())
        while not task.done():
            self._dispatcher_fiber.switch()  # type: ignore
        asyncio._set_running_loop(self._loop)
        return task.result()

    async def close(self) -> None:
        await self.dispose()

    def __repr__(self) -> str:
        return "<Disposable>"



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, List, Literal, Optional

from playwright._impl._api_structures import NameValue

if TYPE_CHECKING:
    from playwright._impl._page import Page


WebStorageKind = Literal["local", "session"]


class WebStorage:
    def __init__(self, page: "Page", kind: WebStorageKind) -> None:
        self._page = page
        self._kind = kind
        self._loop = page._loop
        self._dispatcher_fiber = page._dispatcher_fiber

    async def items(self) -> List[NameValue]:
        result = await self._page._channel.send_return_as_dict(
            "webStorageItems", None, {"kind": self._kind}
        )
        return (result or {}).get("items", [])

    async def get_item(self, name: str) -> Optional[str]:
        result = await self._page._channel.send_return_as_dict(
            "webStorageGetItem", None, {"kind": self._kind, "name": name}
        )
        return (result or {}).get("value")

    async def set_item(self, name: str, value: str) -> None:
        await self._page._channel.send(
            "webStorageSetItem",
            None,
            {"kind": self._kind, "name": name, "value": value},
        )

    async def remove_item(self, name: str) -> None:
        await self._page._channel.send(
            "webStorageRemoveItem", None, {"kind": self._kind, "name": name}
        )

    async def clear(self) -> None:
        await self._page._channel.send("webStorageClear", None, {"kind": self._kind})



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, List

from playwright._impl._api_structures import VirtualCredential
from playwright._impl._helper import locals_to_params

if TYPE_CHECKING:
    from playwright._impl._browser_context import BrowserContext


class Credentials:
    def __init__(self, browser_context: "BrowserContext") -> None:
        self._browser_context = browser_context
        self._loop = browser_context._loop
        self._dispatcher_fiber = browser_context._dispatcher_fiber

    async def install(self) -> None:
        await self._browser_context._channel.send("credentialsInstall", None)

    async def create(
        self,
        rpId: str,
        id: str = None,
        userHandle: str = None,
        privateKey: str = None,
        publicKey: str = None,
    ) -> VirtualCredential:
        result = await self._browser_context._channel.send_return_as_dict(
            "credentialsCreate", None, locals_to_params(locals())
        )
        return (result or {})["credential"]

    async def delete(self, id: str) -> None:
        await self._browser_context._channel.send("credentialsDelete", None, {"id": id})

    async def get(
        self,
        rpId: str = None,
        id: str = None,
    ) -> List[VirtualCredential]:
        result = await self._browser_context._channel.send_return_as_dict(
            "credentialsGet", None, locals_to_params(locals())
        )
        return (result or {}).get("credentials", [])



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from typing import List, Tuple, Union

from playwright._impl._api_structures import FilePayload

FormDataValue = Union[bool, float, str, pathlib.Path, FilePayload]


class FormData:
    def __init__(self) -> None:
        self._fields: List[Tuple[str, FormDataValue]] = []

    def set(self, name: str, value: FormDataValue) -> "FormData":
        self._fields = [(n, v) for (n, v) in self._fields if n != name]
        self._fields.append((name, value))
        return self

    def append(self, name: str, value: FormDataValue) -> "FormData":
        self._fields.append((name, value))
        return self



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from typing import TYPE_CHECKING, Optional, Union

from playwright._impl._artifact import Artifact
from playwright._impl._helper import Error

if TYPE_CHECKING:  # pragma: no cover
    from playwright._impl._page import Page


class Video:
    def __init__(self, page: "Page", artifact: Optional[Artifact]) -> None:
        self._loop = page._loop
        self._dispatcher_fiber = page._dispatcher_fiber
        self._page = page
        self._is_remote = page._connection.is_remote
        self._artifact = artifact

    def __repr__(self) -> str:
        return f"<Video page={self._page}>"

    async def path(self) -> pathlib.Path:
        if self._is_remote:
            raise Error(
                "Path is not available when using browserType.connect(). Use save_as() to save a local copy."
            )
        if not self._artifact:
            raise Error("Video recording has not been started.")
        return self._artifact.absolute_path

    async def save_as(self, path: Union[str, pathlib.Path]) -> None:
        if self._page._connection._is_sync and not self._page._is_closed:
            raise Error(
                "Page is not yet closed. Close the page prior to calling save_as"
            )
        if not self._artifact:
            raise Error("Video recording has not been started.")
        await self._artifact.save_as(path)

    async def delete(self) -> None:
        if self._artifact:
            await self._artifact.delete()



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Python package `playwright` is a Python library to automate Chromium,
Firefox and WebKit with a single API. Playwright is built to enable cross-browser
web automation that is ever-green, capable, reliable and fast.
"""

from typing import Any, Optional, Union, overload

import playwright._impl._api_structures
import playwright._impl._errors
import playwright._impl._form_data
import playwright.async_api._generated
from playwright._impl._assertions import (
    APIResponseAssertions as APIResponseAssertionsImpl,
)
from playwright._impl._assertions import LocatorAssertions as LocatorAssertionsImpl
from playwright._impl._assertions import PageAssertions as PageAssertionsImpl
from playwright.async_api._context_manager import PlaywrightContextManager
from playwright.async_api._generated import (
    APIRequest,
    APIRequestContext,
    APIResponse,
    APIResponseAssertions,
    Browser,
    BrowserContext,
    BrowserType,
    CDPSession,
    ConsoleMessage,
    Dialog,
    Download,
    ElementHandle,
    FileChooser,
    Frame,
    FrameLocator,
    JSHandle,
    Keyboard,
    Locator,
    LocatorAssertions,
    Mouse,
    Page,
    PageAssertions,
    Playwright,
    Request,
    Response,
    Route,
    Selectors,
    Touchscreen,
    Video,
    WebError,
    WebSocket,
    WebSocketRoute,
    Worker,
)

ChromiumBrowserContext = BrowserContext

Cookie = playwright._impl._api_structures.Cookie
FilePayload = playwright._impl._api_structures.FilePayload
FormData = playwright._impl._form_data.FormData
FloatRect = playwright._impl._api_structures.FloatRect
Geolocation = playwright._impl._api_structures.Geolocation
HttpCredentials = playwright._impl._api_structures.HttpCredentials
BrowserBindResult = playwright._impl._api_structures.BrowserBindResult
DebuggerLocation = playwright._impl._api_structures.DebuggerLocation
DebuggerPausedDetails = playwright._impl._api_structures.DebuggerPausedDetails
PdfMargins = playwright._impl._api_structures.PdfMargins
Position = playwright._impl._api_structures.Position
ProxySettings = playwright._impl._api_structures.ProxySettings
ResourceTiming = playwright._impl._api_structures.ResourceTiming
ScreencastFrame = playwright._impl._api_structures.ScreencastFrame
ScreencastSize = playwright._impl._api_structures.ScreencastSize
SourceLocation = playwright._impl._api_structures.SourceLocation
StorageState = playwright._impl._api_structures.StorageState
StorageStateCookie = playwright._impl._api_structures.StorageStateCookie
ViewportSize = playwright._impl._api_structures.ViewportSize
VirtualCredential = playwright._impl._api_structures.VirtualCredential

Error = playwright._impl._errors.Error
TimeoutError = playwright._impl._errors.TimeoutError


def async_playwright() -> PlaywrightContextManager:
    return PlaywrightContextManager()


class Expect:
    _unset: Any = object()

    def __init__(self) -> None:
        self._timeout: Optional[float] = None

    def set_options(self, timeout: Optional[float] = _unset) -> None:
        """
        This method sets global `expect()` options.

        Args:
            timeout (float): Timeout value in milliseconds. Default to 5000 milliseconds.

        Returns:
            None
        """
        if timeout is not self._unset:
            self._timeout = timeout

    @overload
    def __call__(
        self, actual: Page, message: Optional[str] = None
    ) -> PageAssertions: ...

    @overload
    def __call__(
        self, actual: Locator, message: Optional[str] = None
    ) -> LocatorAssertions: ...

    @overload
    def __call__(
        self, actual: APIResponse, message: Optional[str] = None
    ) -> APIResponseAssertions: ...

    def __call__(
        self, actual: Union[Page, Locator, APIResponse], message: Optional[str] = None
    ) -> Union[PageAssertions, LocatorAssertions, APIResponseAssertions]:
        return self._dispatch(actual, message, is_soft=False)

    @overload
    def soft(self, actual: Page, message: Optional[str] = None) -> PageAssertions: ...

    @overload
    def soft(
        self, actual: Locator, message: Optional[str] = None
    ) -> LocatorAssertions: ...

    @overload
    def soft(
        self, actual: APIResponse, message: Optional[str] = None
    ) -> APIResponseAssertions: ...

    def soft(
        self, actual: Union[Page, Locator, APIResponse], message: Optional[str] = None
    ) -> Union[PageAssertions, LocatorAssertions, APIResponseAssertions]:
        """
        Creates a [soft assertion](https://playwright.dev/python/docs/test-assertions#soft-assertions).
        Failing soft assertions do not abort test execution, but mark the test
        as failed. Multiple failures from the same test are surfaced together
        at the end of the test.

        Requires the [pytest-playwright](https://pypi.org/project/pytest-playwright/)
        plugin to establish the per-test scope that collects soft assertion
        failures.
        """
        return self._dispatch(actual, message, is_soft=True)

    def _dispatch(
        self,
        actual: Union[Page, Locator, APIResponse],
        message: Optional[str],
        is_soft: bool,
    ) -> Union[PageAssertions, LocatorAssertions, APIResponseAssertions]:
        if isinstance(actual, Page):
            return PageAssertions(
                PageAssertionsImpl(
                    actual._impl_obj,
                    self._timeout,
                    message=message,
                    is_soft=is_soft,
                )
            )
        elif isinstance(actual, Locator):
            return LocatorAssertions(
                LocatorAssertionsImpl(
                    actual._impl_obj,
                    self._timeout,
                    message=message,
                    is_soft=is_soft,
                )
            )
        elif isinstance(actual, APIResponse):
            return APIResponseAssertions(
                APIResponseAssertionsImpl(
                    actual._impl_obj,
                    self._timeout,
                    message=message,
                    is_soft=is_soft,
                )
            )
        raise ValueError(f"Unsupported type: {type(actual)}")


expect = Expect()


__all__ = [
    "expect",
    "async_playwright",
    "APIRequest",
    "APIRequestContext",
    "APIResponse",
    "Browser",
    "BrowserBindResult",
    "BrowserContext",
    "BrowserType",
    "CDPSession",
    "ChromiumBrowserContext",
    "ConsoleMessage",
    "Cookie",
    "Dialog",
    "Download",
    "ElementHandle",
    "Error",
    "FileChooser",
    "FilePayload",
    "FloatRect",
    "FormData",
    "Frame",
    "FrameLocator",
    "Geolocation",
    "HttpCredentials",
    "JSHandle",
    "Keyboard",
    "Locator",
    "Mouse",
    "DebuggerLocation",
    "DebuggerPausedDetails",
    "Page",
    "PdfMargins",
    "Position",
    "Playwright",
    "ProxySettings",
    "Request",
    "ResourceTiming",
    "Response",
    "Route",
    "ScreencastFrame",
    "ScreencastSize",
    "Selectors",
    "SourceLocation",
    "StorageState",
    "StorageStateCookie",
    "TimeoutError",
    "Touchscreen",
    "Video",
    "ViewportSize",
    "VirtualCredential",
    "WebError",
    "WebSocket",
    "WebSocketRoute",
    "Worker",
]



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import inspect
import math
import uuid
from asyncio.tasks import Task
from typing import Any, Callable, List, Optional, Tuple, Union

from pyee import EventEmitter

from playwright._impl._connection import ChannelOwner
from playwright._impl._errors import Error, TimeoutError


class Waiter:
    def __init__(self, channel_owner: ChannelOwner, event: str) -> None:
        self._result: asyncio.Future = asyncio.Future()
        self._wait_id = uuid.uuid4().hex
        self._loop = channel_owner._loop
        self._pending_tasks: List[Task] = []
        self._channel = channel_owner._channel
        self._registered_listeners: List[Tuple[EventEmitter, str, Callable]] = []
        self._logs: List[str] = []
        self._wait_for_event_info_before(self._wait_id, event)

    def _wait_for_event_info_before(self, wait_id: str, event: str) -> None:
        self._channel.send_no_reply(
            "__waitInfo__",
            None,
            {
                "waitId": wait_id,
                "phase": "before",
                "event": event,
            },
            title=f'Wait for event "{event}"',
        )

    def _wait_for_event_info_after(self, wait_id: str, error: Exception = None) -> None:
        self._channel._connection.wrap_api_call_sync(
            lambda: self._channel.send_no_reply(
                "__waitInfo__",
                None,
                {
                    "waitId": wait_id,
                    "phase": "after",
                    **({"error": str(error)} if error else {}),
                },
            ),
            True,
        )

    def reject_on_event(
        self,
        emitter: EventEmitter,
        event: str,
        error: Union[Error, Callable[..., Error]],
        predicate: Callable = None,
    ) -> None:
        def on_match() -> None:
            self._reject(error() if callable(error) else error)

        def listener(event_data: Any = None) -> None:
            self._evaluate_predicate(predicate, event_data, on_match)

        emitter.on(event, listener)
        self._registered_listeners.append((emitter, event, listener))

    def reject_on_timeout(self, timeout: float, message: str) -> None:
        if timeout == 0:
            return

        async def reject() -> None:
            await asyncio.sleep(timeout / 1000)
            self._reject(TimeoutError(message))

        self._pending_tasks.append(self._loop.create_task(reject()))

    def _cleanup(self) -> None:
        for task in self._pending_tasks:
            if not task.done():
                task.cancel()
        for listener in self._registered_listeners:
            listener[0].remove_listener(listener[1], listener[2])

    def _fulfill(self, result: Any) -> None:
        self._cleanup()
        if not self._result.done():
            self._result.set_result(result)
        self._wait_for_event_info_after(self._wait_id)

    def _reject(self, exception: Exception) -> None:
        self._cleanup()
        if exception:
            base_class = TimeoutError if isinstance(exception, TimeoutError) else Error
            exception = base_class(str(exception) + format_log_recording(self._logs))
        if not self._result.done():
            self._result.set_exception(exception)
        self._wait_for_event_info_after(self._wait_id, exception)

    def wait_for_event(
        self,
        emitter: EventEmitter,
        event: str,
        predicate: Callable = None,
    ) -> None:
        def listener(event_data: Any = None) -> None:
            self._evaluate_predicate(
                predicate, event_data, lambda: self._fulfill(event_data)
            )

        emitter.on(event, listener)
        self._registered_listeners.append((emitter, event, listener))

    def _evaluate_predicate(
        self,
        predicate: Optional[Callable],
        event_data: Any,
        on_match: Callable[[], None],
    ) -> None:
        if predicate is None:
            on_match()
            return
        try:
            result = predicate(event_data)
        except Exception as e:
            self._reject(e)
            return
        if inspect.iscoroutine(result):

            async def _await_predicate(coro: Any) -> None:
                try:
                    matched = await coro
                except Exception as e:
                    self._reject(e)
                    return
                if matched and not self._result.done():
                    on_match()

            self._pending_tasks.append(self._loop.create_task(_await_predicate(result)))
            return
        if result:
            on_match()

    def result(self) -> asyncio.Future:
        return self._result

    def log(self, message: str) -> None:
        self._logs.append(message)
        try:
            self._channel._connection.wrap_api_call_sync(
                lambda: self._channel.send_no_reply(
                    "__waitInfo__",
                    None,
                    {
                        "waitId": self._wait_id,
                        "phase": "log",
                        "message": message,
                    },
                ),
                True,
            )
        except Exception:
            pass


def throw_on_timeout(timeout: float, exception: Exception) -> asyncio.Task:
    async def throw() -> None:
        await asyncio.sleep(timeout / 1000)
        raise exception

    return asyncio.create_task(throw())


def format_log_recording(log: List[str]) -> str:
    if not log:
        return ""
    header = " logs "
    header_length = 60
    left_length = math.floor((header_length - len(header)) / 2)
    right_length = header_length - len(header) - left_length
    new_line = "\n"
    return f"{new_line}{'=' * left_length}{header}{'=' * right_length}{new_line}{new_line.join(log)}{new_line}{'=' * header_length}"



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Literal, Optional, Union

from playwright._impl._api_structures import ScreencastFrame, ScreencastSize
from playwright._impl._artifact import Artifact
from playwright._impl._connection import from_nullable_channel
from playwright._impl._disposable import DisposableStub
from playwright._impl._errors import Error
from playwright._impl._helper import locals_to_params

if TYPE_CHECKING:  # pragma: no cover
    from playwright._impl._page import Page


ScreencastFrameCallback = Callable[[ScreencastFrame], Any]
ScreencastPosition = Literal[
    "bottom",
    "bottom-left",
    "bottom-right",
    "top",
    "top-left",
    "top-right",
]
ScreencastCursor = Literal[
    "none",
    "pointer",
]


class Screencast:
    def __init__(self, page: "Page") -> None:
        self._page = page
        self._loop = page._loop
        self._dispatcher_fiber = page._dispatcher_fiber
        self._started = False
        self._save_path: Optional[Union[str, Path]] = None
        self._on_frame: Optional[ScreencastFrameCallback] = None
        self._artifact: Optional[Artifact] = None
        page._channel.on("screencastFrame", lambda params: self._dispatch_frame(params))

    def _dispatch_frame(self, params: dict) -> None:
        if not self._on_frame:
            return
        data = params["data"]
        if isinstance(data, str):
            data = base64.b64decode(data)
        result = self._on_frame(
            {
                "data": data,
                "timestamp": params.get("timestamp", 0),
                "viewportWidth": params["viewportWidth"],
                "viewportHeight": params["viewportHeight"],
            }
        )
        if hasattr(result, "__await__"):
            self._page._loop.create_task(result)

    async def start(
        self,
        onFrame: ScreencastFrameCallback = None,
        path: Union[str, Path] = None,
        quality: int = None,
        size: ScreencastSize = None,
    ) -> DisposableStub:
        if self._started:
            raise Error("Screencast is already started")
        self._started = True
        self._on_frame = onFrame
        result = await self._page._channel.send_return_as_dict(
            "screencastStart",
            None,
            {
                "size": size,
                "quality": quality,
                "sendFrames": bool(onFrame),
                "record": bool(path),
            },
        )
        artifact_channel = (result or {}).get("artifact")
        if artifact_channel:
            self._artifact = from_nullable_channel(artifact_channel)
            self._save_path = path
        return DisposableStub(lambda: self.stop(), self._page)

    async def stop(self) -> None:
        self._started = False
        self._on_frame = None
        await self._page._channel.send("screencastStop", None)
        if self._save_path and self._artifact:
            await self._artifact.save_as(self._save_path)
        self._artifact = None
        self._save_path = None

    async def show_actions(
        self,
        duration: float = None,
        position: ScreencastPosition = None,
        fontSize: int = None,
        cursor: ScreencastCursor = None,
    ) -> DisposableStub:
        await self._page._channel.send(
            "screencastShowActions", None, locals_to_params(locals())
        )
        return DisposableStub(lambda: self.hide_actions(), self._page)

    async def hide_actions(self) -> None:
        await self._page._channel.send("screencastHideActions", None)

    async def show_overlay(self, html: str, duration: float = None) -> DisposableStub:
        result = await self._page._channel.send_return_as_dict(
            "screencastShowOverlay", None, locals_to_params(locals())
        )
        overlay_id = (result or {}).get("id")
        return DisposableStub(
            lambda: self._page._channel.send(
                "screencastRemoveOverlay", None, {"id": overlay_id}
            ),
            self._page,
        )

    async def show_chapter(
        self,
        title: str,
        description: str = None,
        duration: float = None,
    ) -> None:
        await self._page._channel.send(
            "screencastChapter", None, locals_to_params(locals())
        )

    async def show_overlays(self) -> None:
        await self._page._channel.send(
            "screencastSetOverlayVisible", None, {"visible": True}
        )

    async def hide_overlays(self) -> None:
        await self._page._channel.send(
            "screencastSetOverlayVisible", None, {"visible": False}
        )



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from typing import Any, Dict, Literal, Optional, Pattern, Union, cast

from playwright._impl._api_structures import TracingGroupLocation
from playwright._impl._artifact import Artifact
from playwright._impl._connection import (
    ChannelOwner,
    from_channel,
    from_nullable_channel,
)
from playwright._impl._disposable import DisposableStub
from playwright._impl._helper import Error, locals_to_params


class Tracing(ChannelOwner):
    def __init__(
        self, parent: ChannelOwner, type: str, guid: str, initializer: Dict
    ) -> None:
        super().__init__(parent, type, guid, initializer)
        self._include_sources: bool = False
        self._is_live: bool = False
        self._stacks_id: Optional[str] = None
        self._is_tracing: bool = False
        self._traces_dir: Optional[str] = None
        self._har_id: Optional[str] = None
        self._har_recorders: Dict[str, Dict[str, str]] = {}

    async def start(
        self,
        name: str = None,
        title: str = None,
        snapshots: bool = None,
        screenshots: bool = None,
        sources: bool = None,
        live: bool = None,
    ) -> None:
        params = locals_to_params(locals())
        self._include_sources = bool(sources)
        self._is_live = bool(live)

        await self._channel.send(
            "tracingStart",
            None,
            {
                "name": name,
                "snapshots": snapshots,
                "screenshots": screenshots,
                "live": live,
            },
        )
        trace_name = await self._channel.send(
            "tracingStartChunk", None, {"title": title, "name": name}
        )
        await self._start_collecting_stacks(trace_name)

    async def start_chunk(self, title: str = None, name: str = None) -> None:
        params = locals_to_params(locals())
        trace_name = await self._channel.send("tracingStartChunk", None, params)
        await self._start_collecting_stacks(trace_name)

    async def _start_collecting_stacks(self, trace_name: str) -> None:
        if not self._is_tracing:
            self._is_tracing = True
            self._connection.set_is_tracing(True)
        self._stacks_id = await self._connection.local_utils.tracing_started(
            self._traces_dir, trace_name, self._is_live
        )

    async def stop_chunk(self, path: Union[pathlib.Path, str] = None) -> None:
        await self._do_stop_chunk(path)

    async def stop(self, path: Union[pathlib.Path, str] = None) -> None:
        await self._do_stop_chunk(path)
        await self._channel.send(
            "tracingStop",
            None,
        )

    async def _do_stop_chunk(self, file_path: Union[pathlib.Path, str] = None) -> None:
        self._reset_stack_counter()

        if not file_path:
            # Not interested in any artifacts
            await self._channel.send("tracingStopChunk", None, {"mode": "discard"})
            if self._stacks_id:
                await self._connection.local_utils.trace_discarded(self._stacks_id)
            return

        is_local = not self._connection.is_remote

        if is_local:
            result = await self._channel.send_return_as_dict(
                "tracingStopChunk", None, {"mode": "entries"}
            )
            await self._connection.local_utils.zip(
                {
                    "zipFile": str(file_path),
                    "entries": result["entries"],
                    "stacksId": self._stacks_id,
                    "mode": "write",
                    "includeSources": self._include_sources,
                }
            )
            return

        result = await self._channel.send_return_as_dict(
            "tracingStopChunk",
            None,
            {
                "mode": "archive",
            },
        )

        artifact = cast(
            Optional[Artifact],
            from_nullable_channel(result.get("artifact")),
        )

        # The artifact may be missing if the browser closed while stopping tracing.
        if not artifact:
            if self._stacks_id:
                await self._connection.local_utils.trace_discarded(self._stacks_id)
            return

        # Save trace to the final local file.
        await artifact.save_as(file_path)
        await artifact.delete()

        await self._connection.local_utils.zip(
            {
                "zipFile": str(file_path),
                "entries": [],
                "stacksId": self._stacks_id,
                "mode": "append",
                "includeSources": self._include_sources,
            }
        )

    def _reset_stack_counter(self) -> None:
        if self._is_tracing:
            self._is_tracing = False
            self._connection.set_is_tracing(False)

    async def group(
        self, name: str, location: TracingGroupLocation = None
    ) -> DisposableStub:
        await self._channel.send("tracingGroup", None, locals_to_params(locals()))
        return DisposableStub(lambda: self.group_end(), self)

    async def group_end(self) -> None:
        await self._channel.send(
            "tracingGroupEnd",
            None,
        )

    async def start_har(
        self,
        path: Union[pathlib.Path, str],
        content: Literal["attach", "embed", "omit"] = None,
        mode: Literal["full", "minimal"] = None,
        urlFilter: Union[str, Pattern[str]] = None,
    ) -> DisposableStub:
        if self._har_id:
            raise Error("HAR recording has already been started")
        is_zip = str(path).endswith(".zip")
        default_content: Literal["attach", "embed", "omit"] = (
            "attach" if is_zip else "embed"
        )
        self._har_id = await self._record_into_har(
            har=path,
            page=None,
            url=urlFilter,
            update_content=content or default_content,
            update_mode=mode or "full",
        )
        return DisposableStub(lambda: self.stop_har(), self)

    async def _record_into_har(
        self,
        har: Union[pathlib.Path, str],
        page: Optional[ChannelOwner],
        url: Union[str, Pattern[str]] = None,
        update_content: Literal["attach", "embed", "omit"] = None,
        update_mode: Literal["full", "minimal"] = None,
        resourcesDir: Optional[str] = None,
    ) -> str:
        is_zip = str(har).endswith(".zip")
        url_glob: Optional[str] = None
        url_regex_source: Optional[str] = None
        url_regex_flags: Optional[str] = None
        if isinstance(url, str):
            url_glob = url
        elif url is not None:
            url_regex_source = url.pattern
            url_regex_flags = "".join(
                flag
                for flag, mask in (("i", 2), ("m", 8), ("s", 16))
                if url.flags & mask
            )
        options: Dict[str, object] = {
            "content": update_content or "attach",
            "mode": update_mode or "minimal",
            "harPath": None if is_zip else str(har),
        }
        if url_glob is not None:
            options["urlGlob"] = url_glob
        if url_regex_source is not None:
            options["urlRegexSource"] = url_regex_source
        if url_regex_flags is not None:
            options["urlRegexFlags"] = url_regex_flags
        if resourcesDir is not None:
            options["resourcesDir"] = resourcesDir
        params: Dict[str, Any] = {"options": options}
        if page is not None:
            params["page"] = page._channel
        result = await self._channel.send_return_as_dict("harStart", None, params)
        har_id = result["harId"]
        self._har_recorders[har_id] = {"path": str(har)}
        return har_id

    async def _export_all_hars(self) -> None:
        for har_id in list(self._har_recorders.keys()):
            await self._export_har(har_id)
        self._har_id = None

    async def stop_har(self) -> None:
        har_id = self._har_id
        if not har_id:
            return
        self._har_id = None
        await self._export_har(har_id)

    async def _export_har(self, har_id: str) -> None:
        params = self._har_recorders.pop(har_id, None)
        if not params:
            return
        is_local = not self._connection.is_remote
        is_zip = params["path"].endswith(".zip")

        if is_local:
            result = await self._channel.send_return_as_dict(
                "harExport", None, {"harId": har_id, "mode": "entries"}
            )
            if not is_zip:
                # Server wrote HAR and resources to the user's chosen paths.
                return
            await self._connection.local_utils.zip(
                {
                    "zipFile": params["path"],
                    "entries": result["entries"],
                    "mode": "write",
                    "includeSources": False,
                }
            )
            return

        result = await self._channel.send_return_as_dict(
            "harExport", None, {"harId": har_id, "mode": "archive"}
        )
        artifact = cast(Artifact, from_channel(result["artifact"]))
        if is_zip:
            await artifact.save_as(params["path"])
            await artifact.delete()
            return
        # Uncompressed har is not supported in thin clients
        await artifact.save_as(params["path"] + ".tmp")
        await artifact.delete()
